"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.compile = void 0;
const tslib_1 = require("tslib");
const utils_1 = require("@morjs/utils");
const path_1 = tslib_1.__importDefault(require("path"));
const constants_1 = require("./constants");
const utils_2 = require("./utils");
const PLUGIN_NAME = 'MOR_WEBPACK_COMPILE_PLUGIN';
/**
 * 通过 runner 的单独实例来编译独立分包
 */
function compileIndependentSubpackages(entryBuilder, runner, compiler) {
    const userConfig = (runner.userConfig || {});
    const PLUGIN_NAME_CHILD = PLUGIN_NAME + '_CHILD';
    // 仅在 bundle 模式下 以及 编译目标不是 web 时提供独立分包的编译支持
    // 非 bundle 模式可以利用, 各个小程序 IDE 本身提供的独立分包编译
    if (userConfig.compileMode !== 'bundle' || userConfig.target === 'web')
        return;
    let independentSubpackages = new Map();
    let independentRunners = new Map();
    const originalUserConfig = (0, utils_1.asArray)(runner.config.userConfig).filter((c) => {
        return (c === null || c === void 0 ? void 0 : c.name) === userConfig.name;
    }) || userConfig;
    // 自动更新独立分包, 并清理 mor 缓存
    runner.hooks.beforeBuildEntries.tapPromise(PLUGIN_NAME_CHILD, async () => {
        const keptRunners = new Map();
        for await (const [, sub] of entryBuilder.independentSubpackages) {
            if (independentRunners.has(sub.root)) {
                keptRunners.set(sub.root, independentRunners.get(sub.root));
                independentRunners.delete(sub.root);
            }
            else {
                // 启动分包编译
                // 组装用户配置
                const newRunner = new utils_1.Runner(utils_1.mor, 
                // 定制用户配置
                {
                    ...originalUserConfig,
                    // 独立分包配置
                    ...{
                        compileType: 'subpackage',
                        srcPath: path_1.default.join(userConfig.srcPath, sub.root),
                        srcPaths: userConfig.srcPaths.map((s) => path_1.default.join(s, sub.root)),
                        outputPath: path_1.default.join(userConfig.outputPath, sub.root),
                        // 独立分包不触发集成流程
                        compose: false
                    }
                }, 
                // 定制 context
                {
                    ...Object.fromEntries(runner.context.entries()),
                    [constants_1.INDEPENDENT_SUBPACAKGE_JSON]: sub,
                    [constants_1.CHILD_COMPILER_RUNNER]: true
                });
                keptRunners.set(sub.root, newRunner);
                // 执行 分包编译
                newRunner.hooks.compiler.tap(PLUGIN_NAME_CHILD, function (childCompiler) {
                    childCompiler.hooks.thisCompilation.tap(PLUGIN_NAME_CHILD, () => {
                        utils_1.logger.info(`编译独立分包 ${sub.root} 中 ...`);
                    });
                    childCompiler.hooks.done.tap({
                        name: PLUGIN_NAME_CHILD,
                        stage: Number.MAX_SAFE_INTEGER
                    }, () => {
                        utils_1.logger.info(`独立分包 ${sub.root} 编译完成`);
                    });
                    childCompiler.hooks.failed.tap(PLUGIN_NAME_CHILD, () => {
                        utils_1.logger.error(`独立分包 ${sub.root} 编译失败`);
                    });
                });
                await newRunner.run({
                    name: runner.commandName,
                    args: runner.commandArgs,
                    options: {
                        ...(runner.commandOptions || {}),
                        // 独立分包不触发集成流程
                        compose: false
                    }
                });
            }
        }
        // 这里关闭 runner 实例 并 清理原来的 runner map
        for await (const [, r] of independentRunners) {
            await r.hooks.shutdown.promise(r);
        }
        independentRunners.clear();
        // 保存新的实例列表
        independentRunners = keptRunners;
        // 保存新的独立分包配置
        independentSubpackages = entryBuilder.independentSubpackages;
    });
    // 需要保留分包产物，这里通过定制 CleanPlugin 的 hook 来完成
    compiler.hooks.emit.tap(PLUGIN_NAME_CHILD, function (compilation) {
        if (independentSubpackages.size) {
            const roots = [];
            independentSubpackages.forEach((sub) => roots.push(sub.root));
            const keepDirRegExp = new RegExp(roots.join('|'));
            utils_1.webpack.CleanPlugin.getCompilationHooks(compilation).keep.tap(PLUGIN_NAME_CHILD, (asset) => {
                if (keepDirRegExp.test(asset))
                    return true;
            });
        }
    });
}
/**
 * 提供 worker 编译支持
 */
function compileWorkers(entryBuilder, runner, compiler) {
    const userConfig = runner.userConfig;
    // 仅在 bundle 模式下 以及 编译目标不是 web 且不是分包编译时提供 worker 的编译支持
    if (userConfig.compileMode !== 'bundle' ||
        userConfig.target === 'web' ||
        userConfig.compileType !== 'miniprogram') {
        return;
    }
    const globalObject = userConfig.globalObject;
    const srcPaths = userConfig.srcPaths;
    let workerEnabed = false;
    /**
     * 需要编译的 worker 文件（不带后缀）
     */
    const workerFiles = new Set();
    const PLUGIN_NAME_WORKER = PLUGIN_NAME + '_WORKER';
    runner.hooks.beforeBuildEntries.tap(PLUGIN_NAME_WORKER, (entryBuilder) => {
        var _a;
        workerEnabed = !!((_a = entryBuilder.appJson) === null || _a === void 0 ? void 0 : _a.workers);
    });
    compiler.hooks.normalModuleFactory.tap(PLUGIN_NAME_WORKER, (factory) => {
        factory.hooks.parser
            .for('javascript/auto')
            .tap(PLUGIN_NAME_WORKER, (parser) => parse(parser));
        factory.hooks.parser
            .for('javascript/dynamic')
            .tap(PLUGIN_NAME_WORKER, (parser) => parse(parser));
        factory.hooks.parser
            .for('javascript/esm')
            .tap(PLUGIN_NAME_WORKER, (parser) => parse(parser));
        function parse(parser) {
            if (!workerEnabed)
                return;
            // 从 createWorker 方法中获取文件引用
            // 并替换为正确的文件名称
            function parseWorker(expression) {
                if (expression.type === 'CallExpression') {
                    if (expression.arguments.length < 1)
                        return;
                    const argExpression = expression.arguments[0];
                    const arg = parser.evaluateExpression(argExpression);
                    if (arg && arg.isString() && arg.string) {
                        const fileName = (0, utils_2.pathWithoutExtname)(arg.string);
                        workerFiles.add(fileName);
                        const dep = new utils_1.webpack.dependencies.ConstDependency(JSON.stringify(fileName + '.js'), argExpression.range);
                        dep.loc = argExpression.loc;
                        parser.state.module.addPresentationalDependency(dep);
                    }
                }
            }
            parser.hooks.call
                .for(`${globalObject}.createWorker`)
                .tap(PLUGIN_NAME_WORKER, parseWorker);
            parser.hooks.call.for('createWorker').tap(PLUGIN_NAME_WORKER, parseWorker);
        }
    });
    compiler.hooks.thisCompilation.tap(PLUGIN_NAME_WORKER, async (compilation) => {
        compilation.hooks.additionalAssets.tapPromise(PLUGIN_NAME_WORKER, async () => {
            if (!workerEnabed || !workerFiles.size)
                return;
            for await (const file of workerFiles) {
                const entryName = file + '.js';
                const workerFile = await entryBuilder.tryReachFileByExts(file, entryBuilder.scriptWithConditionalExts, srcPaths);
                const workerCompiler = compilation.createChildCompiler(PLUGIN_NAME_WORKER, {
                    filename: entryName,
                    chunkFilename: entryName,
                    globalObject: 'self'
                });
                workerCompiler.context = compiler.context;
                new utils_1.webpack.webworker.WebWorkerTemplatePlugin().apply(workerCompiler);
                new utils_1.webpack.EntryPlugin(compiler.context, workerFile, entryName).apply(workerCompiler);
                // 确保 hash 生成正确
                workerCompiler.hooks.make.tapAsync(PLUGIN_NAME_WORKER, (childCompilation, callback) => {
                    childCompilation.hooks.afterHash.tap(PLUGIN_NAME_WORKER, () => {
                        childCompilation.hash = compilation.hash;
                        childCompilation.fullHash = compilation.fullHash;
                    });
                    callback();
                });
                // 执行 worker 子编译
                await new Promise((resolve) => {
                    workerCompiler.runAsChild((err, entries, childCompilation) => {
                        if (!err &&
                            childCompilation.errors &&
                            childCompilation.errors.length) {
                            err = childCompilation.errors[0];
                        }
                        // 检查 entry 是否存在
                        const entry = entries &&
                            entries[0] &&
                            entries[0].files.values().next().value;
                        if (!err && !entry) {
                            err = new utils_1.webpack.WebpackError(`编译 worker 失败，未找到文件: ${workerFile}`);
                        }
                        if (err) {
                            compilation.errors.push(err);
                        }
                        resolve(null);
                    });
                });
            }
        });
    });
}
/**
 * 小程序编译逻辑
 */
async function compile(webpackWrapper, entryBuilder, runner) {
    const userConfig = runner.userConfig;
    // 锁定配置
    const config = webpackWrapper.buildConfig();
    /* 入口解析 */
    config.entry = async () => await entryBuilder.build();
    /* 准备 webpack compiler 实例 并触发 compiler hook */
    const { compiler } = webpackWrapper.prepare();
    runner.hooks.compiler.call(compiler);
    // 打印 文件变更
    compiler.hooks.invalid.tap(PLUGIN_NAME, function (fileOrDirPath) {
        if (!(0, utils_2.isChildCompilerRunner)(runner)) {
            utils_1.logger.info(`触发变更: \n=> ${fileOrDirPath}`);
        }
    });
    // 尝试保存 entry 构建的结果
    // entry 缓存仅在 非 default 模式下生效
    if (userConfig.cache === true &&
        userConfig.compileMode === constants_1.CompileModes.bundle) {
        let entryCacheTimeout;
        compiler.hooks.done.tap(PLUGIN_NAME, () => {
            const isWatching = !!webpackWrapper.watching;
            if (entryCacheTimeout)
                clearTimeout(entryCacheTimeout);
            entryCacheTimeout = setTimeout(async () => {
                await entryBuilder.trySaveToCache();
            }, 
            // watching 状态下 延迟 500ms 写入
            isWatching ? 500 : 0);
        });
    }
    // 如果开启了 autoClean 则开始编译前先清空输出目录
    if (userConfig.autoClean) {
        await utils_1.fsExtra.emptyDir(userConfig.outputPath);
    }
    // 独立分包编译支持
    compileIndependentSubpackages(entryBuilder, runner, compiler);
    // workers 编译支持
    compileWorkers(entryBuilder, runner, compiler);
    // 执行 webpack 编译
    await webpackWrapper.run();
}
exports.compile = compile;
//# sourceMappingURL=compile.js.map