"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConfigParserPlugin = void 0;
const tslib_1 = require("tslib");
const utils_1 = require("@morjs/utils");
const path_1 = tslib_1.__importDefault(require("path"));
const compilerPlugins_1 = require("../compilerPlugins");
const EntryBuilderMap = new WeakMap();
/**
 * 多端编译的配置解析和转换
 * 这里仅提供通用的处理, 端的差异由编译插件来解决
 */
class ConfigParserPlugin {
    constructor() {
        this.name = 'ConfigParserPlugin';
    }
    apply(runner) {
        runner.hooks.beforeRun.tap(this.name, () => {
            const userConfig = runner.userConfig;
            // 如果是转换为 web 则不做处理
            if (userConfig.target === 'web')
                return;
            // 是否需要注入全局组件, 判断条件如下
            // 1. 当编译模式为 小程序
            // 2. 目标平台不支持全局组件
            // 3. 全局组件不为空
            let shouldInjectGlobalComponents = userConfig.compileType === utils_1.CompileTypes.miniprogram &&
                !(0, compilerPlugins_1.getComposedCompilerPlugins)().supportGlobalComponents[userConfig.target];
            let globalComponentNames = [];
            runner.hooks.afterBuildEntries.tap(this.name, (entries, entryBuilder) => {
                var _a;
                EntryBuilderMap.set(runner, entryBuilder);
                // 判断是否存在全局组件
                if (shouldInjectGlobalComponents) {
                    globalComponentNames = Object.keys(((_a = entryBuilder.appJson) === null || _a === void 0 ? void 0 : _a.usingComopnents) || {});
                    shouldInjectGlobalComponents = globalComponentNames.length > 0;
                }
                return entries;
            });
            runner.hooks.configParser.tapPromise(this.name, async (config, options) => {
                const entryBuilder = EntryBuilderMap.get(runner);
                const fileInfo = options.fileInfo;
                const isPageOrComponent = fileInfo.entryType === utils_1.EntryType.component ||
                    fileInfo.entryType === utils_1.EntryType.page;
                // 替换app/页面/组件中的路径
                if (fileInfo.entryType === utils_1.EntryType.app ||
                    isPageOrComponent ||
                    fileInfo.entryType === utils_1.EntryType.npmComponent) {
                    // 替换引用为相对路径
                    if (config === null || config === void 0 ? void 0 : config.usingComponents) {
                        for (const key in config.usingComponents) {
                            const referencePath = entryBuilder.getRealReferencePath(fileInfo.path, config.usingComponents[key]);
                            if (referencePath)
                                config.usingComponents[key] = referencePath;
                        }
                        // 为页面和组件注入全局组件
                        // 为节约编译性能，目前的全局组件注入比较粗暴
                        // 直接为每个页面和组件添加全局组件
                        // 没有分析页面的 template 中是否实际有引用
                        if (shouldInjectGlobalComponents && isPageOrComponent) {
                            const entry = entryBuilder.getEntryByFilePath(fileInfo.path);
                            for (const name of globalComponentNames) {
                                // 如果已存在同名组件, 则跳过
                                if (config.usingComponents[name])
                                    continue;
                                const referencePath = entryBuilder.getGlobalComponentRelativePath(entry, name);
                                if (referencePath)
                                    config.usingComponents[name] = referencePath;
                            }
                        }
                    }
                    // 抽象节点的默认组件相对路径替换
                    if (config === null || config === void 0 ? void 0 : config.componentGenerics) {
                        utils_1.lodash.forEach(config === null || config === void 0 ? void 0 : config.componentGenerics, function (val, key) {
                            if (typeof val === 'object' &&
                                typeof (val === null || val === void 0 ? void 0 : val.default) === 'string') {
                                const referencePath = entryBuilder.getRealReferencePath(fileInfo.path, val.default);
                                if (referencePath) {
                                    config.componentGenerics[key].default = referencePath;
                                }
                            }
                        });
                    }
                }
                // plugin.json 的 main 字段补足后缀
                if (userConfig.compileType === utils_1.CompileTypes.plugin &&
                    fileInfo.entryType === utils_1.EntryType.plugin &&
                    fileInfo.entryFileType === utils_1.EntryFileType.config) {
                    if (config.main) {
                        const extname = path_1.default.extname(config.main);
                        if (!extname) {
                            config.main = config.main + '.js';
                        }
                    }
                }
                // 修改 项目配置 中的路径
                // 指定为当前目录
                if (fileInfo.entryType === utils_1.EntryType.project &&
                    fileInfo.entryFileType === utils_1.EntryFileType.config) {
                    if (userConfig.compileType === utils_1.CompileTypes.miniprogram) {
                        config.miniprogramRoot = './';
                    }
                    else if (userConfig.compileType === utils_1.CompileTypes.plugin) {
                        config.pluginRoot = './';
                    }
                    // 这里移除 项目配置中的自定义脚本 避免一些脚本重复执行
                    if (config.scripts) {
                        delete config.scripts;
                    }
                }
                return config;
            });
        });
    }
}
exports.ConfigParserPlugin = ConfigParserPlugin;
//# sourceMappingURL=configParserPlugin.js.map